(in-package :famiclom)

;; This is really just a playground to get a feel for the data layout,
;; write simple name table or pattern table viewers, etc.
;; SEE: http://wiki.nesdev.com/w/index.php/Nametable
;; SEE: http://badderhacksnet.ipage.com/badderhacks/index.php?view=article&id=270

(defvar *framebuffer* (bytevector (* 256 240 4))
  "A framebuffer of bytes to render out. 256x240xRGBA")

(defun ppu-restore (file)
  "Set the PPU state based on a PPU dump file from FCEUX."
  (with-accessors ((ppu nes-ppu)) *nes*
    (let ((image (alexandria:read-file-into-byte-vector
                  (app-path "roms/~a.bin" file))))
      (setf (ppu-pattern-lo ppu) (subseq image #x0000 #x1000)
            (ppu-pattern-hi ppu) (subseq image #x1000 #x2000)
            (ppu-nametable-lo ppu) (subseq image #x2000 #x2400)
            (ppu-palette ppu) (subseq image #x3f00 #x3f20)))))

(defun get-tile-bytes (n &optional (table 'ppu-pattern-hi))
  "Get the bytes comprising the Nth tile in the pattern-table."
  (let ((offset (* #x10 n))
        (pattern-table (funcall table (nes-ppu *nes*))))
    (subseq pattern-table offset (+ offset #x10))))

(defun get-tile-addr (row col &optional (table 'ppu-nametable-lo))
  "Look up the Pattern Table address of the Name Table entry at row, col."
  (let ((offset (+ (* row 32) col))
        (nametable (funcall table (nes-ppu *nes*))))
    (aref nametable offset)))

(defun get-tile (row col)
  "Return a TILE corresponding to the Name Table entry at row, col."
  (let ((tile-addr (get-tile-addr row col)))
    (make-tile :bytes (get-tile-bytes tile-addr))))

(defun get-arena (n)
  "Get the byte that defines the Nth arena."
  (aref (ppu-nametable-lo (nes-ppu *nes*)) (+ #x3c0 n)))

(defun get-arena-blocks (n)
  "Get the 2x2 tile blocks in the Nth Arena."
  (let ((start (multiple-value-bind (row col) (floor n 8)
                 (+ (* 128 row) (* 4 col)))))
    ; Get the 4 tiles in each block, starting from the top left tile, i.
    (loop for i in '(0 2 64 66)
       collect (list (+ start i 00) (+ start i 01)
                     (+ start i 32) (+ start i 33)))))

;; A tile is 8x8, A block is 2x2 tiles, an Arena is 2x2 blocks
(defun draw-bg ()
  (dotimes (i 64)
    (let* ((arena (get-arena i))
           (blocks (get-arena-blocks i)))
      ;; TODO: now what?
      ;; For each block...
      ;; 1. get the palette and pattern-table entries.
      ;; 2. draw the block to the framebuffer.
      ;; After all the arenas are done, draw the framebuffer.
      )))

(defun draw-bg ()
  ;; May want some with-scanline/with-tile control-flow macros here later...
  (dotimes (row 30)
    (dotimes (col 32)
      (let ((tile (get-tile row col))
            ;; PSEUDOCODE!
            (arena (get-arena))
            (color (get-color arena tile)))
        (draw-to-framebuffer tile color))
      ;; TODO: now what?
      ;; 1. determine 'attribute arena' for the tile
      ;; 2. determine palette for the attribute arena
      ;; 3. draw-attribute-arena to the framebuffer.
      ;; 4. after all the tiles are done, draw the framebuffer.
      )))

(defun draw-sprites ()
  ;; I HAVE NO GOD DAMNED IDEA
  )

;; (progn
;;   (draw-bg)
;;   (draw-sprites)
;;   (render-framebuffer))
